use itertools::Itertools;
use std::num::NonZeroU8;

use crate::{board::Board, util::Symb};

/// Returns an iterator of (sort, coords, char_idx, f32) for each empty slot in the listed partials.
/// - sort is the index of this slot.
/// - coords are the coordinate of this slot's partial
/// - char_idx is the index of this slot in its partial
/// - f32 is the influence of this slot
pub fn free_slots_by_influence(board: &Board) -> Option<Vec<(usize, f32)>> {
	// Fill all empty slots with fives and compute starting value
	let filled = {
		// This should always result in an evaluatable expression,
		// since parenthesis do not exist.
		// (the only way to divide by zero is by doing something like /(5-2+3) )
		let f = Board::from_board(board.get_board().map(|x| match x {
			None => Symb::from_char('5'),
			_ => x,
		}));

		f
	};

	let base = filled.evaluate()?;

	// Test each slot:
	// Increase its value by 1, and record its effect on the
	// expression's total value.
	// This isn't a perfect metric, but it's pretty good.
	let mut slots: Vec<(usize, f32)> = board
		.get_board()
		.iter()
		.enumerate()
		.filter_map(|(i, s)| if s.is_some() { None } else { Some(i) })
		.map(|i| {
			let mut new_tree = filled.clone();
			new_tree.get_board_mut()[i] = Some(Symb::from_char('6').unwrap());

			// This shouldn't ever be None
			(i, new_tree.evaluate().unwrap() - base)
		})
		.collect();

	// Sort by most to least influence
	slots.sort_by(|a, b| {
		b.1.abs()
			.partial_cmp(&a.1.abs())
			.unwrap_or(std::cmp::Ordering::Equal)
	});
	Some(slots)
}

/// Find the maximum possible value of the given board
#[allow(dead_code)]
pub fn maximize_value(board: &Board) -> Option<Board> {
	let n_free = board.get_board().iter().filter(|x| x.is_none()).count();

	// Assume we have 10 or fewer available slots
	if n_free >= 10 {
		panic!()
	}

	let available_numbers = (0..=9)
		.map(|x| match x {
			0 => Symb::Zero,
			x => Symb::Number(NonZeroU8::new(x).unwrap()),
		})
		.filter(|x| !board.contains(*x))
		.collect::<Vec<_>>();

	let slots = free_slots_by_influence(&board)?;

	let all_symbols = {
		// We need this many from the bottom, and this many from the top.
		let neg_count = slots.iter().filter(|(_, x)| *x <= 0.0).count();
		let pos_count = slots.iter().filter(|(_, x)| *x > 0.0).count();

		let mut a_iter = available_numbers
			.iter()
			.take(neg_count)
			.chain(available_numbers.iter().rev().take(pos_count).rev());

		let mut g = slots
			// Group slots with equal weights
			// and count the number of elements in each group
			.iter()
			.group_by(|x| x.1)
			.into_iter()
			.map(|(_, x)| x.count())
			// Generate the digits we should try for each group of
			// equal-weight slots
			.map(|s| {
				(0..s)
					.map(|_| a_iter.next().unwrap().clone())
					.permutations(s)
					.unique()
					.collect_vec()
			})
			// Now, covert this to an array of all cartesian products
			// of this set of sets
			.multi_cartesian_product()
			.map(|x| x.iter().flatten().cloned().collect_vec())
			.map(|v| slots.iter().zip(v).collect_vec())
			.collect_vec();

		// Sort these vectors so the order of values
		// matches the order of empty slots
		g.iter_mut()
			.for_each(|v| v.sort_by(|(a, _), (b, _)| b.0.partial_cmp(&a.0).unwrap()));
		g.into_iter()
			.map(|v| v.into_iter().map(|(_, s)| s).collect_vec())
	};

	let mut best_board = None;
	let mut best_value = None;
	for i in all_symbols {
		let mut i_iter = i.iter();
		let filled = Board::from_board(board.get_board().map(|x| match x {
			None => i_iter.next().cloned(),
			_ => x,
		}));

		let val = filled.evaluate();

		if let Some(val) = val {
			if let Some(best) = best_value {
				if val > best {
					best_value = Some(val);
					best_board = Some(filled)
				}
			} else {
				best_value = Some(val);
				best_board = Some(filled)
			}
		}
	}

	best_board
}
